/////////////////////////////////////////////////////////////////////////////
///
/// @file confinfo.h
///
/// Writing configuration settings to an output stream.
/// This file contains the definition of a function which gathers
/// various configuration parameters and outputs them to a stream
/// in a nice text format.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on April 10, 2008. Last revision: September 15, 2010.


#ifndef _CMGRAPHS_CONFINFO_H_
#define _CMGRAPHS_CONFINFO_H_


// include some standard C++ header files
#include <ostream>

// include local header files
#include "config.h"
#include "typedefs.h"


// --------------------------------------------------
// ---------------- SHOW CONFIG INFO ----------------
// --------------------------------------------------

/// Writes the configuration information to the given output stream.
/// Precedes each line with the prefix provided.
/// If the array "maxIter" is given then includes the information
/// on the chunks of the parameter space that will be sent to workers.
inline void showConfigInfo (std::ostream &f, const char *prefix,
	const parCoord *maxIter)
{
	if (!prefix)
		prefix = "";

	f << prefix << "=================================\n";
	f << prefix << "Configuration of the computations\n";
	f << prefix << "=================================\n";
	f << prefix << "\n";

	f << prefix << "THE PHASE SPACE:\n";
	f << prefix << "dim = " << spaceDim << "\n";
	f << prefix << "depth: from " << initialDepth << " to " <<
		finalDepth << "\n";

	f << prefix << "offset:";
	for (int i = 0; i < spaceDim; ++ i)
		f << (i ? ", " : " ") << spaceOffset [i];
	f << "\n";

	f << prefix << "width:";
	for (int i = 0; i < spaceDim; ++ i)
		f << (i ? " x " : " ") << spaceWidth [i];
	f << "\n";

	f << prefix << "size in cubes:";
	for (int i = 0; i < spaceDim; ++ i)
		f << (i ? " x " : " ") << (1 << finalDepth);
	f << "\n";
	f << prefix << "\n";

	f << prefix << "THE PARAMETER SPACE:\n";
	f << prefix << "dim = " << paramDim << "\n";

	f << prefix << "subdivisions:";
	for (int i = 0; i < paramDim; ++ i)
		f << (i ? " x " : " ") << paramSubdiv [i];
	f << "\n";

	if (maxIter)
	{
		f << prefix << "no. of tasks:";
		int nTasks = 1;
		for (int i = 0; i < paramDim; ++ i)
		{
			f << (i ? " x " : " ") << maxIter [i];
			nTasks *= maxIter [i];
		}
		f << " = " << nTasks << "\n";
	}

	f << prefix << "selected parameters (see below):";
	for (int i = 0; i < paramDim; ++ i)
		f << (i ? ", " : " ") << paramSelect [i];
	f << "\n";

	f << prefix << "total number of parameters: " <<
		paramCount << "\n";

	f << prefix << "ranges:";
	for (int i = 0; i < paramCount; ++ i)
		f << (i ? " x " : " ") << "[" << paramLeft [i] << "," <<
			paramRight [i] << "]";
	f << "\n";
	f << prefix << "\n";

	f << prefix << "LIMITATIONS:\n";
	f << prefix << "max refine depth = " << refineDepth << "\n";
	f << prefix << "max refine size 0 = " << maxRefineSize0 << "\n";
	f << prefix << "max refine size 1 = " << maxRefineSize1 << "\n";
	f << prefix << "max image diameter = " << maxImageDiameter << "\n";
	f << prefix << "max image volume = " << maxImageVolume << "\n";
	f << prefix << "max index pair size = " << maxIndexPairSize << "\n";
	f << prefix << "\n";

	f << prefix << "OTHER PARAMETERS:\n";
	f << prefix << "max join size = " << maxJoinSize << "\n";
	f << prefix << "max join connection = " << maxJoinConnection << "\n";
	f << prefix << "max join distance = " << maxJoinDistance << "\n";
	f << prefix << "compare Morse ordering = " <<
		compareMorseOrdering << "\n";
	f << prefix << "ignore isol. for cont. = " <<
		ignoreIsolationForContinuation << "\n";
	f << prefix << "\n";

	f << prefix << "COMPILATION-DEPENDENT SETTINGS:\n";
	f << prefix << "sizeof (spcCoord) = " << sizeof (spcCoord) << "\n";
	f << prefix << "sizeof (parCoord) = " << sizeof (parCoord) << "\n";
	f << prefix << "sizeof (int_t) = " << sizeof (int_t) << "\n";
	f << prefix << "sizeof (size_t) = " << sizeof (size_t) << "\n";
	f << prefix << "sizeof (short) = " << sizeof (short) << "\n";
	f << prefix << "sizeof (int) = " << sizeof (int) << "\n";
	f << prefix << "sizeof (long) = " << sizeof (long) << "\n";
	f << prefix << "sizeof (int *) = " << sizeof (int *) << "\n";
	f << prefix << "DimBits = " << chomp::homology::DimBits << "\n";
	f << prefix << "MaxBasDim = " << chomp::homology::MaxBasDim << "\n";
	f << prefix << "max no. of cubes = " <<
		chomp::homology::NumMask << "\n";
	f << prefix << "\n";

	f << prefix << "=================================\n";
	return;
} /* showConfigInfo */


#endif // _CMGRAPHS_CONFINFO_H_

