/////////////////////////////////////////////////////////////////////////////
///
/// @file m_lorenz_a.h
///
/// An approximate time-t map for the Lorenz equations.
/// This map defines a sample time-t map for an ODE.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package. This is free software:
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this software; see the file "license.txt". If not,
// please, see <https://www.gnu.org/licenses/>.

// Started on June 26, 2012. Last revision: June 26, 2012.


#ifndef _CMGRAPHS_LORENZAPPR_H_
#define _CMGRAPHS_LORENZAPPR_H_


// include local header files
#include "maptype.h"
#include "odeapprox.h"


// --------------------------------------------------
// ------------ the Lorenz time-t map ------------
// --------------------------------------------------

/// An approximate time-t map for the Lorenz equations.
class MapLorenzApprox: public MapOdeApprox
{
public:
	/// The constructor of a Lorenz map object.
	MapLorenzApprox ();

private:
	/// The vector field.
	void vectorField (double, const double *x, double *y, int dim) const;

}; /* class MapLorenzApprox */

// --------------------------------------------------

inline MapLorenzApprox::MapLorenzApprox ():
	MapOdeApprox (0.05/20, 20, 4)
{
	return;
} /* MapLorenzApprox::MapLorenzApprox */

inline void MapLorenzApprox::vectorField (double,
	const double *x, double *y, int dim) const
{
	if (dim != 3)
		throw "This model is valid in dimension 3 only.";

	// is this the Lorenz equation with reversed time?
	const bool reversed = false;

	// retrieve the parameters (the left bound only)
	double sigma = getLeftParam (0);
	double rho = getLeftParam (1);
	double beta = getLeftParam (2);

	// compute the vector field
	y [0] = sigma * (x [1] - x [0]);
	y [1] = x [0] * (rho - x [2]) - x [1];
	y [2] = x [0] * x [1] - beta * x [2];

	// reverse the vector field if requested to
	if (reversed)
	{
		for (int i = 0; i < dim; ++ i)
			y [i] = -y [i];
	}

	return;
} /* MapLorenzApprox::vectorField */


#endif // _CMGRAPHS_LORENZAPPR_H_

